package net.sf.itcb.common.client.aop;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;

import javax.annotation.Resource;
import javax.xml.transform.Source;
import javax.xml.transform.dom.DOMResult;

import net.sf.itcb.common.business.exceptions.BusinessItcbException;
import net.sf.itcb.common.business.exceptions.ExceptionMappingErrors;
import net.sf.itcb.common.client.exceptions.ClientItcbException;
import net.sf.itcb.common.client.exceptions.ClientItcbExceptionMappingErrors;
import net.sf.itcb.common.client.security.provider.SecurementCredentialsProvider;

import org.aopalliance.intercept.MethodInterceptor;
import org.aopalliance.intercept.MethodInvocation;
import org.springframework.beans.factory.annotation.Configurable;
import org.springframework.context.i18n.LocaleContextHolder;
import org.springframework.security.core.userdetails.UserDetails;
import org.springframework.ws.client.core.support.WebServiceGatewaySupport;
import org.springframework.ws.soap.SoapElement;
import org.springframework.ws.soap.SoapFaultDetailElement;
import org.springframework.ws.soap.client.SoapFaultClientException;
import org.springframework.ws.soap.saaj.SaajSoapFaultException;
import org.springframework.ws.soap.security.wss4j.Wss4jSecurityInterceptor;
import org.w3c.dom.DOMException;

/**
 * This interceptor allows to use generic client in order to call a spring webservice.
 * It can manage each interface which respects those rules :
 * <ul>
 * <li>Only one parameter <b>*Request</b> generated by JAXB using an XSD file. This object contains all information needed by the service</li>
 * <li>A return object <b>*Response</b> generated by JAXB using an XSD file</li>
 * </ul>
 * If it doesn't match, a {@link net.sf.itcb.common.client.exceptions.ClientItcbException} is thrown with {@link net.sf.itcb.common.client.exceptions.ClientExceptionMappingErrors#}.
 *
 * If a {@link org.springframework.ws.soap.security.wss4j.Wss4jSecurityInterceptor} interceptor is declared in the <b>webServiceTemplate</b> property, 
 * then the <b>securementCredentialsProvider</b> (@see net.sf.itcb.common.client.security.provider.SecurementCredentialsProvider.SecurementCredentialsProvider} property has to be defined too.
 * If not, a {@link net.sf.itcb.common.client.exceptions.ClientItcbException.ClientItcbException} is thrown with {@link net.sf.itcb.common.client.exceptions.ClientExceptionMappingErrors#COMMON_CLIENT_NOT_SET_SECUREMENT_PROVIDER}.
 * 
 * If the server has thrown a BusinessItcbException (or a child), the ClientInterceptor knows how to resolve it, then a BusinessItcbException is thrown in client applications. 
 *  
 * @author Pierre Le Roux
 */
@Configurable(preConstruction=true)
public class ClientInterceptor extends WebServiceGatewaySupport implements MethodInterceptor {
	
	
	@Resource(name="common-clientResources")
	protected org.springframework.context.MessageSource messageSource;
	
	private SecurementCredentialsProvider securementCredentialsProvider;
	
	public void setSecurementCredentialsProvider(SecurementCredentialsProvider securementCredentialsProvider) {
		this.securementCredentialsProvider= securementCredentialsProvider;
	}
	
	public Object invoke(MethodInvocation arg0) throws Throwable {
		for (int i = 0; i < getWebServiceTemplate().getInterceptors().length; i++) {
			if(getWebServiceTemplate().getInterceptors()[i] instanceof Wss4jSecurityInterceptor) {
				if(securementCredentialsProvider == null) {
					throw new ClientItcbException(ClientItcbExceptionMappingErrors.COMMON_CLIENT_NOT_SET_SECUREMENT_PROVIDER, messageSource.getMessage("common-client.exception.not_set_securement_provider", new Object[] {arg0.getMethod()}, LocaleContextHolder.getLocale()));
				}
				UserDetails user = securementCredentialsProvider.getUserCredentials();
				Wss4jSecurityInterceptor securityInterceptor = (Wss4jSecurityInterceptor)getWebServiceTemplate().getInterceptors()[i];
				if(user!= null) {
					securityInterceptor.setSecurementUsername(user.getUsername());
					securityInterceptor.setSecurementPassword(user.getPassword());
				}
			}
		};
		if(arg0.getArguments().length != 1) {
			throw new ClientItcbException(ClientItcbExceptionMappingErrors.COMMON_CLIENT_NOT_CORRECT_INTERFACE, messageSource.getMessage("common-client.exception.not_correct_interface", new Object[] {arg0.getMethod()}, LocaleContextHolder.getLocale()));
		}
		try  {
			return getWebServiceTemplate().marshalSendAndReceive(arg0.getArguments()[0]); 
		}
		catch (SoapFaultClientException sfce) {
			if(sfce.getFaultCode().getNamespaceURI().equals(BusinessItcbException.NAMESPACE_URI) ) {
				makeBusinessItcbException(sfce, arg0);
			}
			throw new ClientItcbException(ClientItcbExceptionMappingErrors.COMMON_CLIENT_TECHNICAL_ERROR, messageSource.getMessage("common-client.exception.technical_error", new Object[] {arg0.getMethod(), getWebServiceTemplate().getDestinationProvider().getDestination()}, LocaleContextHolder.getLocale()));
		}
		// toute autre erreur qu'une erreur
		catch (Throwable technicalException) {
			throw new ClientItcbException(ClientItcbExceptionMappingErrors.COMMON_CLIENT_TECHNICAL_ERROR, messageSource.getMessage("common-client.exception.technical_error", new Object[] {arg0.getMethod(), getWebServiceTemplate().getDestinationProvider().getDestination()}, LocaleContextHolder.getLocale()));
		}
		
	}

	private void makeBusinessItcbException(SoapFaultClientException sfce, MethodInvocation arg0) throws ClientItcbException, BusinessItcbException {
		try {
			while(sfce.getSoapFault().getFaultDetail().getDetailEntries().hasNext()) {
				SoapFaultDetailElement saajSoapFaultDetailElement = sfce.getSoapFault().getFaultDetail().getDetailEntries().next(); 
				if(saajSoapFaultDetailElement.getName().getLocalPart().equals(BusinessItcbException.EXCEPTION_CLASS.getLocalPart())) {
					if(saajSoapFaultDetailElement.getResult() instanceof DOMResult) {
						String exceptionClassName = ((DOMResult)saajSoapFaultDetailElement.getResult()).getNode().getTextContent();
						Class<?> exceptionClass = Class.forName(exceptionClassName);
						Constructor<?>[] constructors= exceptionClass.getConstructors();
						for (Constructor<?> constructor : constructors) {
							if(constructor.getParameterTypes().length == 2) {
								Class<?> classErrorMapping = constructor.getParameterTypes()[0];
								BusinessItcbException bie = (BusinessItcbException)constructor.newInstance( classErrorMapping.getMethod("valueOf", String.class).invoke(classErrorMapping, sfce.getFaultCode().getLocalPart()) , sfce.getLocalizedMessage());
								throw bie;
							}
						}
					}
				}
			}
		}
		catch(BusinessItcbException bie) {
			throw bie;
		}
		// No mapping found
		catch(Throwable t) {
			throw new ClientItcbException(ClientItcbExceptionMappingErrors.COMMON_CLIENT_TECHNICAL_ERROR, messageSource.getMessage("common-client.exception.technical_error", new Object[] {arg0.getMethod(), getWebServiceTemplate().getDestinationProvider().getDestination()}, LocaleContextHolder.getLocale()));
	
		}
	}
	
}
